# $Id: aselect.py 4232 2011-03-11 14:20:53Z wsl $
# $URL: https://svn.non-gnu.uvt.nl/uvt-dev/tags/sources/meresco/bzv-9.9.4/src/aselect.py $

import pprint
import re
from meresco.core import Observable
from socket import socket, error as sockerr
from weightless import compose
from cgi import parse_qs
from urllib import quote as urlencode, unquote_plus as urldecode, urlencode as unparse_q
from urlparse import urlsplit
from weightless import Suspend
from traceback import print_stack, print_exc

_cookie_re = re.compile('[^;=\s]+\s*=\s*(?:"[^"]*"|[^;\s]*)')

def _unparse_qs(q):
	for key, val in q.iteritems():
		yield urlencode(key) + '=' + urlencode(val)

def unparse_qs(q):
	return '&'.join(_unparse_qs(q))

def parse_qs(q):
	res = {}
	for s in q.split('&'):
		key, sep, val = s.partition('=')
		res[urldecode(key)] = urldecode(val)
	return res

def plakurl(url, q):
	if url.find('?') < 0:
		sep = '?'
	else:
		sep = '&'
	return url + sep + unparse_qs(q)

def dumpert(*args):
	pp = pprint.PrettyPrinter(indent=2)
	print pp.pformat(*args)

def spy(iter):
	for s in iter:
		print s,
		yield s

# betere opzet zou misschien zijn om eerst te controleren welke actie nodig
# is (if any) en die actie dan te yielden, verpakt in een Suspend-object.

class AselectClient(object):
	def __init__(self, infra, appid, base, force, req):
		self.infra = infra
		self.appid = appid
		self.base = base
		self.force = force
		self.reactor = None
		self.suspend = None
		self.request = req
		self.ok = False
		self.done = False
		self.uid = None
		self.redirect = None
		self.cookies = {}
		self.handler = compose(self._handler())
		self.this = None

	def __call__(self):
		# called by reactor (for each iteration)
		try:
			self.handler.next()
		except StopIteration:
			self.ok = True
			self.suspend.resume()
		except Exception, e:
			self.suspend.throw(e)

	def check(self):
		try:
			self.handler.next()
		except StopIteration:
			# nothing to do
			self.ok = True
		else:
			# showtime
			s = Suspend(self.handle)
			yield s
			s.getResult()

	def handle(self, suspend):
		# called once by Suspend
		self.suspend = suspend
		self.reactor = suspend._reactor
		self()

	def _connect(self):
		sok = socket()
		sok.setblocking(0)
		self.sok = sok
		try:
			sok.connect(("127.0.0.1", 1495))
		except sockerr, (errno, string):
			if errno == 115:
				self.reactor.addWriter(sok, self)
				yield
				self.reactor.removeWriter(sok)
			else:
				raise

	def _send(self, str):
		sok = self.sok
		self.reactor.addWriter(sok, self)
		while len(str):
			yield
			r = sok.send(str)
			str = str[r:]
		self.reactor.removeWriter(sok)

	def _getline(self, res):
		sok = self.sok
		self.reactor.addReader(sok, self)
		str = ''
		while len(str) < 4096:
			yield
			r = sok.recv(4096)
			if not r:
				break
			off = r.find("\n")
			if off != -1:
				str = str + r[:off]
				break
			str = str + r
		self.reactor.removeReader(sok)
		res.append(str)

	def _comm(self, res, **kwargs):
		yield self._connect()
		yield self._send(unparse_qs(kwargs) + "\n")
		lines = []
		yield self._getline(lines)
		res.update(parse_qs(lines[0]))
		self.sok.close()

	def _url(self):
		baseURL = self.base
		path = self.request['path']
		arguments = self.request['arguments']
		if arguments:
			return ''.join([baseURL, path, '?', unparse_q(arguments, True)])
		else:
			return baseURL + path

	def _bakecookie(self, key, val):
		scheme, host, path, query, fragment = urlsplit(self.base)
		if path == '':
			path = '/'
		yield urlencode(key)
		yield "="
		yield urlencode(val)
		yield "; Path="
		yield path
		if scheme == 'https':
			yield "; Secure"
		if host != 'localhost':
			yield "; Host="
			yield host

	def _cookiejar(self):
		res = {}
		headers = self.request['Headers']
		header = headers.get('Cookie')
		if header is not None:
			for s in _cookie_re.findall(header):
				key, sep, val = s.partition('=')
				res[urldecode(key)] = urldecode(val)
		return res

	def _verify_ticket(self):
		jar = self._cookiejar()
		cookie = jar.get('aselect')
		if cookie is None:
			return
		aselect = parse_qs(cookie)
		if not self.reactor:
			yield # showtime
		res = {}
		yield self._comm(res, request = 'verify_ticket', **aselect)
		if res['result_code'] == '0000':
			self.uid = aselect['uid']
			self.done = True
		else:
			self.infra.logger.info('verify_ticket fail: ' + res['result_code'])

	def _verify_credentials(self):
		arguments = self.request['arguments']
		credentials = arguments.get('aselect_credentials', [])
		if len(credentials) == 0:
			return
		if not self.reactor:
			yield # showtime
		res = {}
		yield self._comm(res,
				request = 'verify_credentials',
				rid = arguments['rid'][0],
				aselect_credentials = credentials[0],
			)
		if res['result_code'] == '0000':
			cookie = {
					'uid': res['uid'],
					'organization': res['organization'],
					'ticket': res['ticket']
				}
			del arguments['rid']
			del arguments['aselect_credentials']
			del arguments['a-select-server']
			self.cookies['aselect'] = unparse_qs(cookie)
			self.redirect = self._url()
			self.done = True
		else:
			self.infra.logger.info('verify_credentials fail: ' + res['result_code'])

	def _authenticate(self):
		if not self.force:
			return
		if not self.reactor:
			yield # showtime
		res = {}
		yield self._comm(res,
				request = 'authenticate',
				app_id = self.appid,
				app_url = self._url()
			)
		if res['result_code'] == '0000':
			url = res['as_url']
			del res['result_code']
			del res['as_url']
			self.redirect = plakurl(url, res)
			self.done = True
		else:
			self.infra.logger.error('authenticate fail: ' + res['result_code'])

	def _handler(self):
		yield self._verify_ticket()
		if not self.done:
			yield self._verify_credentials()
		if not self.done:
			yield self._authenticate()

	def response(self):
		yield "HTTP/1.1 302 Go here first\n"
		yield "Pragma: no-cache\n"
		yield "Cache-Control: no-cache, must-revalidate\n"
		for key, val in self.cookies.iteritems():
			yield "Set-Cookie: "
			yield self._bakecookie(key, val)
			yield "\n"
		if self.redirect:
			yield "Location: "
			yield self.redirect
			yield "\n"
		yield "Content-Type: text/html\n\n"
		yield "<html><body><p>An error occurred</p></body></html>\n"

class Aselect(Observable):
	def __init__(self, infra, regex, baseURL, appid):
		Observable.__init__(self)
		self.infra = infra
		self.re = re.compile(regex)
		self.base = baseURL
		self.appid = appid

	def _krak(self):
		yield "HTTP/1.1 500 Internal server error\n"
		yield "Pragma: no-cache"
		yield "Cache-Control: no-cache, must-revalidate\n"
		yield "Content-Type: text/html\n\n"
		yield "<html><body><p>Please try again later</p></body></html>\n"

	def handleRequest(self, *args, **kwargs):
		if self.re.match(kwargs['path']):
			#dumpert(kwargs)
			force = 'login' in kwargs.get('arguments', {}).get('aselect', [])
			login = AselectClient(self.infra, self.appid, self.base, force, kwargs)
			yield login.check()
			if login.ok:
				if login.uid:
					kwargs['uid'] = login.uid
					yield self.all.handleRequest(*args, **kwargs)
				elif force:
					yield login.response()
				else:
					yield self.all.handleRequest(*args, **kwargs)
			else:
				yield self._krak()
		else:
			yield self.all.handleRequest(*args, **kwargs)
